/* CHMtools v0.1 */
/* Copyright 2001 Matthew T. Russotto */
/*
    This file is part of CHMtools

    CHMtools is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CHMtools is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <string.h>
#include "chmlib.h"
#include "fixendi.h"
#include "lzx.h"

#define FILELEN_HSECT 0
#define DIR_HSECT 1
#define CONTENT_FORMAT "::DataSpace/Storage/%s/Content"
#define CONTROLDATA_FORMAT "::DataSpace/Storage/%s/ControlData"
#define SPANINFO_FORMAT "::DataSpace/Storage/%s/SpanInfo"
#define LIST_FORMAT "::DataSpace/Storage/%s/Transform/List"
#define INSTANCEDATA_FORMAT "::DataSpace/Storage/%s/Transform/%s/InstanceData/"
#define RT_FORMAT "::DataSpace/Storage/%s/Transform/%s/InstanceData/ResetTable"
#define TRANSFORM_FORMAT "::Transform/%s/"
#define NAMELIST "::DataSpace/NameList"

//#define DEBUG
#ifdef DEBUG
#define DPRINTF fprintf
#else
#define DPRINTF while (0) fprintf
#endif

//!!! Hobo
//#define USE_BIGCACHE

#ifdef DEBUG
static void
get_guid(ubyte *buf, guid_t *guid)
{
    memcpy(guid, buf, sizeof(guid_t));
    FIXENDIAN32(guid->guid1);
    FIXENDIAN16(guid->guid2[0]);
    FIXENDIAN16(guid->guid2[1]);
}

static void
make_guid_string(guid_t *guid, char *s)
{
    sprintf(s, "{%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X}",
        guid->guid1, guid->guid2[0], guid->guid2[1],
        guid->guid3[0], guid->guid3[1], guid->guid3[2], guid->guid3[3],
        guid->guid3[4], guid->guid3[5], guid->guid3[6], guid->guid3[7]);
}
#endif

static void guid_fix_endian(guid_t *guid)
{
    FIXENDIAN32(guid->guid1);
    FIXENDIAN16(guid->guid2[0]);
    FIXENDIAN16(guid->guid2[1]);
}

static int readheader(chmfile *c)
{
    int     ret;

    if(fseek(c->cf, 0, SEEK_SET)) return 1; //!!! Hobo
    ret = fread(&c->ch, 1, sizeof(chmheader), c->cf);
    if(ret != sizeof(chmheader)) return 1; //!!! Hobo
    if(c->ch.version != 3) DPRINTF(stdout, "WARNING: CHMTools only deals with version 3\n");
    guid_fix_endian(&c->ch.unk_guid1);
    guid_fix_endian(&c->ch.unk_guid2);
    FIXENDIAN32(c->ch.tot_hdrlen);
    if(fseek(c->cf, 0x58, SEEK_SET)) return 1; //!!! Hobo
    ret = fread(&c->content_offset, 1, sizeof(c->content_offset), c->cf);
    if(ret != sizeof(c->content_offset)) return 1; //!!! Hobo
    FIXENDIAN32(c->content_offset);

#ifdef DEBUG
    {
        char    str[40];

        DPRINTF(stdout, "itsf: %-4.4s\n", c->ch.itsf);
        DPRINTF(stdout, "tot_hdrlen: %08x\n", c->ch.tot_hdrlen);
        DPRINTF(stdout, "content_offset: %08x\n", c->content_offset);
        make_guid_string(&c->ch.unk_guid1, str);
        DPRINTF(stdout, "guid1: %s\n", str);
        make_guid_string(&c->ch.unk_guid2, str);
        DPRINTF(stdout, "guid2: %s\n", str);
    }
#endif

    if(fseek(c->cf, sizeof(chmheader), SEEK_SET)) return 1; //!!! Hobo
    return ret;
}

static int readhsectable(chmfile *c)
{
    int     i;
    int     result;
    int     nhsecs = 2;

    result = fread(c->hs, sizeof(hsecentry), nhsecs, c->cf);
    if(result != nhsecs) return 1; //!!! Hobo
    for (i = 0; i < nhsecs; i++) {
        FIXENDIAN32(c->hs[i].offset);
        FIXENDIAN32(c->hs[i].length);
    }
    return result;
}

static int readdirheader(chmfile *c)
{
    int     result;

    if(fseek(c->cf, c->hs[DIR_HSECT].offset, SEEK_SET)) return 1; //!!! Hobo
    result = fread(&c->dh, 1, sizeof(dirheader), c->cf);
    if(result != sizeof(dirheader)) return 1; //!!! Hobo
    FIXENDIAN32(c->dh.chunksize);
    FIXENDIAN32(c->dh.indexchunk);
    FIXENDIAN32(c->dh.ndirchunks);
    FIXENDIAN32(c->dh.firstpmglchunk);
    FIXENDIAN32(c->dh.lastpmglchunk);
    FIXENDIAN32(c->dh.length);
    guid_fix_endian(&c->dh.unk_guid1);
#ifdef DEBUG
    {
        char    str[40];

        DPRINTF(stdout, "directory header length: %x\n", c->dh.length);
        DPRINTF(stdout, "chunksize: %x\n", c->dh.chunksize);
        DPRINTF(stdout, "ndirchunks: %x\n", c->dh.ndirchunks);
        DPRINTF(stdout, "indexchunk: %x\n", c->dh.indexchunk);
        DPRINTF(stdout, "firstpmglchunk: %x\n", c->dh.firstpmglchunk);
        DPRINTF(stdout, "lastpmglchunk: %x\n", c->dh.lastpmglchunk);
        make_guid_string(&c->dh.unk_guid1, str);
        DPRINTF(stdout, "guid1: %s\n", str);
    }
#endif
    return result;
}

static ulong getencint(ubyte **p)
{
    ulong accum = 0;

    while ((**p) & 0x80) {
        accum = (accum << 7) | ((*(*p)++) & 0x7F);
    }
    accum = (accum << 7) | (*(*p)++);
    return accum;
}

static int
read_chm_dir(chmfile *c)
{
    int                 length;
    dirheader           dh;
    ubyte               *buf;
    ubyte               *bufend;
    ubyte               *p, *oldp;
    int                 namelen;
    int                 i;
    unsigned long       section, offset, dlength;
    int                 nentries;
    int                 nchunks;
    pmglchunkheader     pmglch;
    fpos_t              chunkstart;
    size_t              rc;

    readdirheader(c);
    chunkstart = ftell(c->cf);
    length = c->dh.chunksize * c->dh.ndirchunks;
    p = buf = malloc(length);
    if(!p) {
        DPRINTF(stdout, "Can not malloc (%u)\n", __LINE__);
        return 1; //!!! Hobo
    }
    bufend = buf;
    nchunks = 0;
    pmglch.next_chunk = 0;
    do {
        if(fseek(c->cf, chunkstart + pmglch.next_chunk * c->dh.chunksize, SEEK_SET)) return 1; //!!! Hobo
        rc = fread(&pmglch, 1, sizeof(pmglch), c->cf);
        if(rc != sizeof(pmglch)) return 1; //!!! Hobo
        FIXENDIAN32(pmglch.next_chunk);
        FIXENDIAN32(pmglch.prev_chunk);
        FIXENDIAN32(pmglch.quickreflen);
    #ifdef DEBUG
        DPRINTF(stdout, "prev_chunk: %d\n", pmglch.prev_chunk);
        DPRINTF(stdout, "next_chunk: %d\n", pmglch.next_chunk);
        DPRINTF(stdout, "quickreflen: %x\n", pmglch.quickreflen);
    #endif
        length = fread(bufend, 1, c->dh.chunksize - sizeof(pmglch) - pmglch.quickreflen, c->cf);
        if(length != c->dh.chunksize - sizeof(pmglch) - pmglch.quickreflen) return 1; //!!! Hobo
        bufend += length;
        nchunks++;
    } while ((pmglch.next_chunk != -1) && (nchunks < c->dh.ndirchunks));

    nentries = 0;

    while (p < bufend) {
        nentries++;
        //namelen = *p++; //!!! ERROR, see next line -- right
        namelen = getencint(&p);
        p += namelen;
        section = getencint(&p);
        offset = getencint(&p);
        dlength = getencint(&p);
    }

#ifdef DEBUG
    DPRINTF(stdout, "nentries (calculated): %#x (%d)\n", nentries, nentries);
#endif

    c->dir = (chm_dir *) malloc(sizeof(chm_dir) + (nentries-1) * sizeof(direntry));
    if(!c->dir) {
        DPRINTF(stdout, "Can not malloc (%u)\n", __LINE__);
        return 1; //!!! Hobo
    }
    p = buf;
    for (i = 0; i < nentries; i++) {
        oldp = p;
        memset(&c->dir->entry[i], 0, sizeof(direntry));
        //namelen = *p++; //!!! ERROR, see next line -- right
        namelen = getencint(&p);

        //!!! next line -- change of getting memory for name
        c->dir->entry[i].name = malloc(namelen + 1);
        if(!c->dir->entry[i].name) {
            DPRINTF(stdout, "Can not malloc (%u)\n", __LINE__);
            return 1; //!!! Hobo
        }

        memcpy(c->dir->entry[i].name, p, namelen);
        c->dir->entry[i].name[namelen] = 0;
        p += namelen;
        c->dir->entry[i].section = getencint(&p);
        offset = getencint(&p);
        dlength = getencint(&p);
        c->dir->entry[i].offset = offset;
        c->dir->entry[i].length = dlength;
    }
#ifdef DEBUG
//    DPRINTF(stdout, "section  offset   length   name\n");
//    for (i = 0; i < nentries; i++)
//        DPRINTF(stdout, "%08lx %08lx %08lx %s\n", c->dir->entry[i].section,
//            c->dir->entry[i].offset, c->dir->entry[i].length,
//            c->dir->entry[i].name);
#endif
    c->dir->nentries = nentries;
    free(buf);
    return 0;
}

static direntry *getdirentry(char *name, chm_dir *dir)
{
    int i;

    for(i = 0; i < dir->nentries; i++)
        if(!strcmp(name, dir->entry[i].name))
            return &dir->entry[i];
    return NULL;
}

int
chm_getfile(chmfile *c, char *name, ulong *length, ubyte **outbuf)
{
    int         i, j;
    int         section;
    int         offset = c->content_offset;
    direntry    *de;
    FILE        *addf;

    *length = 0;
    *outbuf = NULL;
    de = getdirentry(name, c->dir);
    //DPRINTF(stdout, "Getting %s, o: %#x, de = %08x %s\n", name, de->offset, de, de ? /*de->name*/ "OK" : "FAIL");
    if (!de) return -1; /* file not found in directory */
    section = de->section;

    if(!c->cs || !c->cs->entry[section].iscompressed) {
        /*
         * either file in section 0 or section is not compressed
         */
        if(c->cs)
            offset += c->cs->entry[section].offset;

        if(fseek(c->cf, de->offset + offset, SEEK_SET)) return -12; //!!! Hobo
        *length = de->length;
        if(!(*outbuf = malloc(*length))) {
            DPRINTF(stdout, "Can not malloc outbuf (%u)\n", __LINE__);
            //return -1;
        }
        else {
            i = fread(*outbuf, 1, *length, c->cf);
            if(i != *length) return -11; //!!! Hobo
        }
    }
#ifndef USE_BIGCACHE
    else {
        /*
         * unpack each file separately
         */
        contsect_xtra   *csx;
        contsecentry    *cse;

        if(!c->cs->entry[section].xdata) {
            /* get extra section info */
            char            fname[2048];
            ulong           flength;
            ubyte           *cfile;
            ulong           window_size, LZX_ri, ver;
            ulong           encnt;
            ullong          uclen, cclen, blksz;
            int             i;

            DPRINTF(stdout, "section: %u, name: %s, offset: %#x\n",
                section, c->cs->entry[section].name, c->cs->entry[section].offset);

            /* get controldata */
            sprintf(fname, CONTROLDATA_FORMAT, c->cs->entry[section].name);
            DPRINTF(stdout, "CD file: %s\n", fname);
            chm_getfile(c, fname, &flength, &cfile);
            if(!cfile) return -2;
            if (memcmp(cfile+4, "LZXC", 4)) {
                DPRINTF(stdout, "Compression method not LZXC: %-4.4s\n", cfile+4);
                free(cfile);
                return -3;
            }
            ver = *(ulong *) (cfile+0x8);
            LZX_ri = *(ulong *) (cfile+0x0c);
            if(ver > 1) LZX_ri *= 0x8000;
            window_size = *(ulong *) (cfile+0x10);
            free(cfile);

            FIXENDIAN32(LZX_ri);
            FIXENDIAN32(window_size);
            switch(window_size) {
                case 1:     window_size = 15; break;
                case 2:     window_size = 16; break;
                case 4:     window_size = 17; break;
                case 8:     window_size = 18; break;
                case 0x10:  window_size = 19; break;
                case 0x20:  window_size = 20; break;
                case 0x40:  window_size = 21; break;
                default:
                DPRINTF(stdout, "Window size invalid: %x\n", window_size);
                return -4;
            }
            DPRINTF(stdout, "CD: window_size: %#x (%u), LZX ri: %#x (%u)\n", window_size, window_size, LZX_ri, LZX_ri);

            /* get ResetTable */
            /* hardcoded string because transform list is broken */
            sprintf(fname, RT_FORMAT, c->cs->entry[section].name, "{7FC28940-9D31-11D0-9B27-00A0C91E9C7C}");
            DPRINTF(stdout, "RT file: %s\n", fname);
            chm_getfile(c, fname, &flength, &cfile);
            if(!cfile) return -6;

            encnt = *(ulong *) (cfile+0x4);
            uclen = *(ullong *) (cfile+0x10);
            cclen = *(ullong *) (cfile+0x18);
            blksz = *(ullong *) (cfile+0x20);
            DPRINTF(stdout, "RT: encnt: %u, uclen: %#x (%u), cclen: %#x (%u), blksz: %#x (%u)\n",
                encnt, (ulong) uclen, (ulong) uclen, (ulong) cclen, (ulong) cclen, (ulong) blksz, (ulong) blksz);

            csx = malloc(sizeof(contsect_xtra) + sizeof(ullong) * (encnt - 1));
            if(!csx) {
                DPRINTF(stdout, "Cannot malloc csx (%u)\n", __LINE__);
                return -7;
            }
            csx->cdofs  = c->content_offset + c->cs->entry[section].offset;
            csx->LZXri  = LZX_ri;
            csx->wsz    = window_size;
            csx->encnt  = encnt;
            csx->ulen   = uclen;
            csx->clen   = cclen;
            csx->blsz   = blksz;
            memcpy(csx->rtofs, cfile+0x28, encnt * sizeof(ullong));
            free(cfile);
            DPRINTF(stdout, "ResetTable offsets:\n");
            for(i = 0; i < csx->encnt; ++i) DPRINTF(stdout, "%08x\n", (ulong) csx->rtofs[i]);
            DPRINTF(stdout, "Compressed data offset: %#x\n", csx->cdofs);

            c->cs->entry[section].xdata = csx;
        }
        /* get file data in compressed section */
        /* There is one change from LZX as defined by Microsoft: After each
         * LZX reset interval (defined in the ControlData file, but in
         * practice equal to the window size) of compressed data is processed,
         * the LZX state is fully reset, as if an entirely new file was being
         * encoded. This allows semi-random access to the compressed data; you
         * can start reading on any reset interval boundary using the reset
         * interval size (chm_ctldata.ri) and the reset table. */
        {
            ubyte   *cdata, *udata, *obuf;
            int     rti;    /* reset table index */
            int     owp;    /* output window position */
            int     rc, to_read, to_deco, written;
            size_t  readres;

            //DPRINTF(stdout, "Decopressing: %s\n", de->name);

            cse = c->cs->entry + section;
            csx = c->cs->entry[section].xdata;

            rti = (de->offset / csx->LZXri) * (csx->LZXri / csx->blsz);
            owp = (de->offset % csx->LZXri);
            DPRINTF(stdout, "rti: %u, owp: %i\n", rti, owp);
            if(!(cdata = malloc(csx->blsz + 1024))) {
                DPRINTF(stdout, "Can not malloc cdata (%u)\n", __LINE__);
            }
            if(!(udata = malloc(csx->blsz + 1024))) {
                DPRINTF(stdout, "Can not malloc udata (%u)\n", __LINE__);
            }
            if(!(obuf = malloc(de->length))) {
                DPRINTF(stdout, "Can not malloc outbuf (%u)\n", __LINE__);
            }

            written = 0;
            rc = LZXinit(csx->wsz);
            if(rc) DPRINTF(stdout, "LZXinit(): %u ", rc);
            do {
                if(rti == csx->encnt - 1) {
                    to_read = csx->clen - csx->rtofs[rti];
                    to_deco = csx->ulen % csx->blsz;
                }
                else {
                    to_read = csx->rtofs[rti+1] - csx->rtofs[rti];
                    to_deco = csx->blsz;
                }
                if(fseek(c->cf, csx->cdofs + csx->rtofs[rti], SEEK_SET)) return -12; //!!! Hobo
                readres = fread(cdata, 1, to_read, c->cf);
                if(readres != to_read) return -11; //!!! Hobo
                rc = LZXdecompress(cdata, udata, to_read, to_deco);
                if(rc) DPRINTF(stdout, "LZXdecompress(): %u\n", rc);
                if(to_deco < owp) {
                    owp -= to_deco;
                    //DPRINTF(stdout, "1: owp: %u\n", owp);
                }
                else {
                    if(to_deco - owp > de->length - written) {
                        memcpy(obuf+written, udata+owp, de->length - written);
                        written = de->length;
                        //DPRINTF(stdout, "2: owp: %u\n", owp);
                    }
                    else {
                        memcpy(obuf+written, udata+owp, to_deco-owp);
                        written += to_deco-owp;
                        owp = 0;
                        //DPRINTF(stdout, "3: owp: %u\n", owp);
                    }
                }
                ++rti;
            } while(written < de->length);

            if(cdata) free(cdata);
            if(udata) free(udata);
            if(rc) return 1;
            *length = de->length;
            *outbuf = obuf;
        }
    }
#else
    else if(c->cs->entry[section].cache) {
        /*
         * contents of sections is unpacked fully in BIG CACHE
         */
        *length = de->length;
        if(!(*outbuf = malloc(*length))) {
            DPRINTF(stdout, "Can not malloc outbuf (%u)\n", __LINE__);
        }
        else {
            memcpy(*outbuf, c->cs->entry[section].cache + de->offset, *length);
        }
    }
    else {
        /*
         * get Content file in section 0 and unpack it into BIG CACHE
         */
        char    fname[4096];
        char    guid_str[80];
        ubyte   *lbp;
        ulong   flength;
        ubyte   *rtfile;
        ubyte   *cdfile;
        ubyte   *cbp;
        ubyte   *contbuf;
        ubyte   *secbuf;
        ulong   rtindex;
        ulong   uclength, clength;
        ulong   contlength;
        ulong   rtlength;
        ulong   window_size;
        guid_t  guid;
        int     result;

        sprintf(fname, CONTROLDATA_FORMAT, c->cs->entry[section].name);

        /* get controldata */
        chm_getfile(c, fname, &flength, &cdfile);
        if (!cdfile)
            return -2;
        cbp = cdfile;
        if (memcmp(cbp+4, "LZXC", 4)) {
            DPRINTF(stdout, "Compression method not LZXC: %-4.4s\n", cbp+4);
            free(cdfile);
            return -3;
        }
        window_size = *(ulong *)(cbp+0x10);
        FIXENDIAN32(window_size);
        free(cdfile);
        switch(window_size) {
            case 1:     window_size = 15; break;
            case 2:     window_size = 16; break;
            case 4:     window_size = 17; break;
            case 8:     window_size = 18; break;
            case 0x10:  window_size = 19; break;
            case 0x20:  window_size = 20; break;
            case 0x40:  window_size = 21; break;
            default:
                DPRINTF(stdout, "Window size invalid: %x\n", window_size);
                return -4;
        }
        DPRINTF(stdout, "CL: window_size: %#x (%u)\n", window_size, window_size);
        strcpy(guid_str, "{7FC28940-9D31-11D0-9B27-00A0C91E9C7C}");
        /* hardcoded string because transform list is broken */
        sprintf(fname, RT_FORMAT, c->cs->entry[section].name, guid_str);
        DPRINTF(stdout, "RTFILE: %s\n", fname);
        chm_getfile(c, fname, &rtlength, &rtfile);
        if(rtfile) {
            uclength = (rtfile[0x10]) | (rtfile[0x11] << 8) | (rtfile[0x12] << 16) | (rtfile[0x13] << 24);
            clength  = (rtfile[0x18]) | (rtfile[0x19] << 8) | (rtfile[0x1a] << 16) | (rtfile[0x1b] << 24);
            DPRINTF(stdout, "uclength = %x (%d), clength = %x (%d)\n", uclength, uclength, clength, clength);
            sprintf(fname, CONTENT_FORMAT, c->cs->entry[section].name);
            chm_getfile(c, fname, &contlength, &contbuf);
            if (clength != contlength)
                DPRINTF(stdout,
                    "Warning: Content Length not same as Compressed Length (without padding) %d %d \n",
                    contlength, clength);

            // !!!Hobo: added +6144 from CAB_INPUTMAX in lzx_int.h
            DPRINTF(stdout, "CL: Getting %u bytes, window_size: %u\n", uclength+6144, window_size);
            secbuf = malloc(uclength+6144);
            if(!secbuf) {
                DPRINTF(stdout, "Can not malloc (%u)\n", __LINE__);
                return -1;
            }
            /* uncompress it */
            LZXinit(window_size);
            result = LZXdecompress(contbuf, secbuf, clength, uclength);
            DPRINTF(stdout, "LZXResult: %d\n", result);
        #ifdef DEBUG_LZX
            if(result != 0) {
                FILE    *fp;
                if(fp = fopen("__contbuf.bin", "wb")) { fwrite(contbuf, clength, 1, fp); fclose(fp); }
                if(fp = fopen("__secbuf.bin", "wb"))  { fwrite(secbuf, uclength, 1, fp); fclose(fp); }
            }
        #endif
            free(contbuf);
            free(rtfile);
            if (result != 0) {
                free(secbuf);
                return -5 + result;
            }
            /* and get the file we want out of it */
            DPRINTF(stdout, "offset = %x\n", de->offset);
            DPRINTF(stdout, "length = %x (%d)\n", de->length, de->length);
            *length = de->length;
            *outbuf = malloc(*length);
            if(!*outbuf) {
                DPRINTF(stdout, "Can not malloc (%u)\n", __LINE__);
                return -1;
            }
            memcpy(*outbuf, secbuf + de->offset, *length);
            c->cs->entry[section].cache = secbuf;
            c->cs->entry[section].cachesize = uclength;
            //          free(secbuf);//!!!
        }
        else return -6;
    }
#endif

    return 0;
}

static int readcontsecs(chmfile *c)
{
    ulong   length;
    ubyte   *buf;
    ubyte   *bufptr;
    int     nentries;
    int     nmlen;
    int     i, j;
    char    secname[4096];

    chm_getfile(c, NAMELIST, &length, &buf);
    nentries = buf[2] | (buf[3] << 8);
    c->cs = (contsecs *) malloc(sizeof(contsecs) + (nentries-1) * sizeof(contsecentry));
    if(!c->cs) {
        DPRINTF(stdout, "Can not malloc (%u)\n", __LINE__);
        return 1;
    }
    c->cs->nentries = nentries;
    bufptr = buf + 4;
    for (i = 0; i < nentries; i++) {
        c->cs->entry[i].xdata = NULL; //!!! Hobo
        nmlen = bufptr[0] | (bufptr[1] << 8);
        bufptr += 2;
        for (j = 0; j <= nmlen; j++) {
            /* this is the lazy way to do wide characters. Since the string
             is pretty much always "MSCompressed" or "Uncompressed", it'll do */
            c->cs->entry[i].name[j] = *bufptr;
            bufptr += 2;
        }
        DPRINTF(stdout, "Section name = %s\n", c->cs->entry[i].name);
        if (!strcmp(c->cs->entry[i].name, "MSCompressed")) {
            c->cs->entry[i].iscompressed = 1;
            /* this is the wrong way to figure out if a file is compressed.
               The real way is to examine the transform.  But the transform
               list is corrupt in most (all?) CHM files, apparently because
               the length was calculated in characters whereas the guid
               is recorded as wide characters */
        }
        else {
            c->cs->entry[i].iscompressed = 0;
        }
        c->cs->entry[i].cache = NULL;
        sprintf(secname, CONTENT_FORMAT, c->cs->entry[i].name);
        c->cs->entry[i].offset = 0;
        if (i == 0) continue;
        for (j = 0; j < c->dir->nentries; j++) {
            if (!strcmp(secname, c->dir->entry[j].name)) {
                c->cs->entry[i].offset = c->dir->entry[j].offset;
            }
        }
    }
    free(buf);
    return 0;
}

chmfile *chm_openfile(char *fname)
{
    chmfile *result;
    FILE *f;

    f = fopen(fname, "rb");
    if (!f) return NULL;
    result = calloc(1, sizeof(chmfile));
    if(!result) {
        fclose(f);
        DPRINTF(stdout, "Can not calloc (%u)\n", __LINE__);
        return NULL;
    }
    result->cf = f;
    readheader(result);
    if(strncmp("ITSF", result->ch.itsf, 4)) { /* !!! Hobo -- test is this CHM signature? */
        DPRINTF(stdout, "Bad signature\n");
        free(result);
        return NULL;
    }
    else DPRINTF(stdout, "Signature Ok.\n");
    readhsectable(result);
    read_chm_dir(result);
    readcontsecs(result);
    return result;
}

void
chm_close(chmfile *c)
{
    int i;

    fclose(c->cf);
    for (i = 0; i < c->cs->nentries; i++) {
        if (c->cs->entry[i].cache) free(c->cs->entry[i].cache);
        if (c->cs->entry[i].xdata) free(c->cs->entry[i].xdata); //!!! Hobo
    }
    free(c->cs);
    for(i = 0; i < c->dir->nentries; ++i) free(c->dir->entry[i].name); //!!! Hobo
    free(c->dir);
}
