#if defined ( _WIN32 )
    #include <windows.h>
#else
    #error Tested only for Win32
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/utime.h>
#include "chmlib.h"
#include "wcxhead.h"
#include "chmdir.h"


//#define DEBUG
#ifdef DEBUG
#   define DPRINTF  fprintf
#   define DOUT     stdout
#else
#   define DPRINTF  while(0) fprintf
#   define DOUT     stderr
#endif


extern int show_spec_files = 0; /* set to 1 if you want to see special CHM files */
extern int show_stor_files = 0; /* set to 1 if you want to see special IStorage files */
extern int use_file_attrib = 1; /* set to 1 if you want attribs for special files */

static tChangeVolProc   WC_ChangeVolProc = NULL;
static tProcessDataProc WC_ProcessDataProc = NULL;


typedef struct CHM_INF {
    char    fn[MAX_PATH];   /* archive file name */
    int     mode;           /* open mode */
    chmfile *cf;
    int     wcx_time;       /* time of chm file in semantic of WCX Writer's Reference */
    time_t  chm_time;       /* time of chm file in semantic of time_t */
    int     ei;             /* current entry index */
} CHM_INF_t;


static char *chm_fnames[] = {
    "/#ITBITS", "/#SYSTEM", "/#IDXHDR", "/#STRINGS", "/#TOCIDX", "/#TOPICS",
    "/#URLSTR", "/#URLTBL", "/#IVB", "/#SUBSETS", "/#WINDOWS",
    "/$FIftiMain", "/$OBJINST", "/$TitleMap",
    "/$WWAssociativeLinks/BTree", "/$WWAssociativeLinks/Data",
    "/$WWAssociativeLinks/Map", "/$WWAssociativeLinks/Property",
    "/$WWKeywordLinks/BTree", "/$WWKeywordLinks/Data",
    "/$WWKeywordLinks/Map", "/$WWKeywordLinks/Property",
    "/#KEY_DATA", "/#KEY_DELETED"
};

static int chm_fnames_cnt = sizeof(chm_fnames) / sizeof(char *);

static int is_chm_file(char *fname)
{
    int     i, rc = 0;

    for(i = 0, rc = 0; i < chm_fnames_cnt && !rc; ++i)
        rc = !strcmp(fname, chm_fnames[i]);
    return rc;
}

static int find_next_entry(struct CHM_INF *ci)  /* in chm dir */
{
    chmfile *cf = ci->cf;
    int found = 0, namelen, i;

    if(ci->ei >= cf->dir->nentries) return 0;

    for(i = ci->ei + 1; i < ci->cf->dir->nentries && !found; ++i) {
        //DPRINTF(DOUT, "find_next_entry(): ce %d: %s\n", i, cf->dir->entry[i].name);
        namelen = strlen(cf->dir->entry[i].name);
        found =  cf->dir->entry[i].name[namelen-1] != '/';          /* no slash at end of name */
        if(!show_stor_files)
            found = found && cf->dir->entry[i].name[0] == '/';      /* slash at start */
        if(!show_spec_files)
            found = found && !is_chm_file(cf->dir->entry[i].name);
    }
    if(found) ci->ei = i - 1;
    return found;
}

static char *convert_chm_fname(char *buf, char *chm_fname)
/* Converts UTF8 encoded chm_fname to system file name (buf)
 * Returns buf
 */
{
    wchar_t     wbuf[2*MAX_PATH];
    char        *s;

    if(*chm_fname == '/') decode_UTF8(wbuf, chm_fname+1);
    else decode_UTF8(wbuf, chm_fname);
    WideCharToMultiByte(CP_ACP, 0, wbuf, -1, buf, MAX_PATH, "_", NULL); // "_" NOT TESTED !!!
    for(s = buf; *s; ++s) {
        if(*s == '/') *s = '\\';
        if(strchr(":*?\"<>", *s)) *s = '_';
    }
    return buf;
}

extern HANDLE __stdcall OpenArchive(tOpenArchiveData *ad)
{
    time_t          ct;
    struct _stat    stb;
    struct tm       *st;
    struct CHM_INF  *ci = NULL;

    //DPRINTF(DOUT, "OpenArchive(): %s.\n", ad->ArcName);
    if(!(ci = calloc(1, sizeof(struct CHM_INF)))) {
        ad->OpenResult = E_NO_MEMORY;
        return 0;
    }
    if(!(ci->cf = chm_openfile(ad->ArcName))) {
        free(ci);
        ad->OpenResult = E_EOPEN;
        return 0;
    }
    /* time */
    if(_stat(ad->ArcName, &stb) == 0) ct = stb.st_mtime;
    else ct = time(NULL);
    st = localtime(&ct);
    ci->wcx_time = (st->tm_year + 1900 - 1980) << 25 | (st->tm_mon + 1) << 21 | st->tm_mday << 16
            | st->tm_hour << 11 | st->tm_min << 5 | st->tm_sec / 2;
    ci->chm_time = ct;
    /* other */
    strcpy(ci->fn, ad->ArcName);
    ci->mode = ad->OpenMode;
    ci->ei = -1;

    DPRINTF(DOUT, "OpenArchive(): entries: %d.\n", ci->cf->dir->nentries);

    if(ci->cf && ci->fn) {
        ad->OpenResult = 0;
        return ci;
    }
    else {
        chm_close(ci->cf);
        free(ci);
        ad->OpenResult = E_UNKNOWN_FORMAT;
        return NULL;
    }
}


extern int __stdcall CloseArchive(HANDLE had)
{
    struct CHM_INF *ci = (struct CHM_INF *) had;

    DPRINTF(DOUT, "CloseArchive(): %s\n", ci->fn);
    chm_close(ci->cf);
    free(ci->cf);
    free(ci);
    return 0;
}

extern int __stdcall ReadHeader(HANDLE had, tHeaderData *hd)
{
    struct CHM_INF  *ci = (struct CHM_INF *) had;
    chmfile         *cf = ci->cf;
    char            *s;
    int             eoa = 0;

    //DPRINTF(DOUT, "ReadHeader(): %s ", hd->ArcName);
    if(!find_next_entry(ci)) {
        eoa = E_END_ARCHIVE;
    }
    else {
        memset(hd, 0, sizeof(*hd));
        strcpy(hd->ArcName, ci->fn);
        s = cf->dir->entry[ci->ei].name;
        if(use_file_attrib) {
            if(is_chm_file(s)) hd->FileAttr = 0x2;
            if(!strncmp(s, "::", 2)) hd->FileAttr = 0x4;
        }
        convert_chm_fname(hd->FileName, s);
        hd->PackSize = hd->UnpSize = cf->dir->entry[ci->ei].length;
        hd->FileTime = ci->wcx_time;
    }
    //DPRINTF(DOUT, "%3d: %s %s\n", ci->ei, hd->FileName, eoa ? "- EOA" : "");
    return eoa;
}


extern int __stdcall ProcessFile(HANDLE had, int op, char *destpath, char *destname)
{
    struct CHM_INF  *ci = (struct CHM_INF *) had;
    chmfile         *cf = ci->cf;
    struct _utimbuf utb;
    int             rc = 0;
    unsigned long   clen;
    unsigned char   *cbuf = NULL;
    FILE            *fp;
    char            fpath[MAX_PATH] = "", *s;
    char            buf[256];
    wchar_t         wbuf[2*MAX_PATH];

//    DPRINTF(DOUT, "ProcessFile(): %s, p: %s, n: %s\n",
//        op == PK_EXTRACT ? "X" : op == PK_TEST ? "T" : op == PK_SKIP ? "S" : "???", destpath, destname);

    switch (op)
    {
        case PK_EXTRACT:
        case PK_TEST:
        /* Either DestName contains the full path and file name and DestPath is NULL,
         * or DestName contains only the file name and DestPath the file path.
         */
        if(destpath && destname) {
            strcpy(fpath, destpath);
            s = fpath + strlen(fpath) - 1;
            if(*s != ':' && *s != '\\') strcat(s, "\\");
            strcat(fpath, destname);
        }
        else if(!destpath && destname) strcpy(fpath, destname);
        else if(destpath && !destname) strcpy(fpath, destpath);
        else convert_chm_fname(fpath, cf->dir->entry[ci->ei].name);

        rc = chm_getfile(cf, cf->dir->entry[ci->ei].name, &clen, &cbuf);
        if(rc) {
            rc = E_BAD_ARCHIVE;
        }
        else {
            if(op == PK_EXTRACT) {
                unsigned int    fattr = FILE_ATTRIBUTE_NORMAL;

                if(use_file_attrib) {
                    s = cf->dir->entry[ci->ei].name;
                    if(is_chm_file(s)) fattr = FILE_ATTRIBUTE_HIDDEN;
                    if(!strncmp(s, "::", 2)) fattr = FILE_ATTRIBUTE_SYSTEM;
                }

                if(!(fp = fopen(fpath, "wb"))) {
                    rc = E_ECREATE;
                }
                else {
                    if(fwrite(cbuf, 1, clen, fp) != clen) rc = E_EWRITE;
                    if(fclose(fp)) rc = E_EWRITE;
                    utb.actime = ci->chm_time;
                    utb.modtime = ci->chm_time;
                    _utime(fpath, &utb);
                    SetFileAttributes(fpath, fattr);
                }
            }
        }
        free(cbuf);
        if(!rc && WC_ProcessDataProc)
            if(!WC_ProcessDataProc(fpath, cf->dir->entry[ci->ei].length))
                rc = E_EABORTED;
        break;

        case PK_SKIP:
        default: rc = 0; break;
    }
    return rc;
}

extern void chmdir_init(HINSTANCE hDLL)
{
#ifdef DEBUG
    SYSTEMTIME ltime;
#endif
    char    dll_name[MAX_PATH+1], *s;

#ifdef DEBUG
    GetLocalTime(&ltime);
    DPRINTF(DOUT, "chmdir_init(): CHMDir.wcx started at %02d:%02d:%02d.\n", ltime.wHour, ltime.wMinute, ltime.wSecond);
#endif
    GetModuleFileName(hDLL, dll_name, MAX_PATH);
    DPRINTF(DOUT, "lib_entry: dll name: %s\n", dll_name);
    for(s = dll_name + strlen(dll_name) - 1; s > dll_name && *s != '\\'; --s) ;
    if(*s == '\\') *s = 0;
    strcat(dll_name, "\\CHMDir.ini");
    DPRINTF(DOUT, "lib_entry: ini name: %s\n", dll_name);
    show_spec_files = GetPrivateProfileInt("Options", "show_spec_files", 0, dll_name);
    show_stor_files = GetPrivateProfileInt("Options", "show_stor_files", 0, dll_name);
    use_file_attrib = GetPrivateProfileInt("Options", "use_file_attrib", 0, dll_name);
    DPRINTF(DOUT, "CHMdir: show_spec_files: %d\n", show_spec_files);
    DPRINTF(DOUT, "CHMdir: show_stor_files: %d\n", show_stor_files);
    DPRINTF(DOUT, "CHMdir: use_file_attrib: %d\n", use_file_attrib);
}

#ifdef CHMDIR_DLL
extern BOOL WINAPI DllMain(HINSTANCE hDLL, DWORD Reason, LPVOID Reserved)
{
    switch(Reason)
    {
        case DLL_PROCESS_ATTACH:
        /* The DLL is being loaded for the first time by a given process.
         Perform per-process initialization here.  If the initialization
         is successful, return TRUE; if unsuccessful, return FALSE. */
        chmdir_init(hDLL);
        break;

    }
    return TRUE;
}
#endif

extern int __stdcall GetPackerCaps(void)
{
    return PK_CAPS_MULTIPLE + PK_CAPS_BY_CONTENT + PK_CAPS_SEARCHTEXT + PK_CAPS_HIDE;
}

extern int __stdcall CanYouHandleThisFile(char *FileName)
{
    FILE    *fp;
    char    buf[16];

    if(!(fp = fopen(FileName, "rb"))) return 0;
    fread(buf, 1, 4, fp);
    fclose(fp);
    return !strncmp(buf, "ITSF", 4);
}

extern void __stdcall SetChangeVolProc(HANDLE had, tChangeVolProc pChangeVolProc1)
{
    DPRINTF(DOUT, "SetChangeVolProc():\n");
    WC_ChangeVolProc = pChangeVolProc1;
}

extern void __stdcall SetProcessDataProc(HANDLE had, tProcessDataProc pProcessDataProc)
{
    DPRINTF(DOUT, "SetProcessDataProc():\n");
    WC_ProcessDataProc = pProcessDataProc;
}
